#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <ctype.h>
#include <pthread.h>
#include <glib.h>

#include "gvar.h"
#include "network.h"
#include "hub_to_hub.h"
#include "main.h"
#include "cnx_lst.h"

#ifdef __CYGWIN__
#define MSG_DONTWAIT 0
#endif

typedef struct
{
	char *rhub_addr;
	unsigned short rhub_port;
	time_t last_checked;
	int is_up;									/* 0=no, 1=yes */
	struct sockaddr_in laddr;				/* only valid when is_up==1 */
	int laddr_len;								/* only valid when is_up==1 */
} RHUB_ENTRY;

GArray *rhub_array=NULL;
G_LOCK_DEFINE(rhub_array);

static int multihub_com_socket=-1;

static void load_rhub_array(void)
{
	char *vl;

	G_LOCK(conf_file);
	G_LOCK(rhub_array);

	vl=e_db_str_get(conf_file,"MHUBLIST");
	if(vl!=NULL)
	{
		gchar **lst;
		int i;

		lst=g_strsplit(vl,"$",0);

		i=0;
		while(lst[i]!=NULL)
		{
			char *t;
			RHUB_ENTRY nw;

			t=strchr(lst[i],':');
			if(t==NULL)
			{
				nw.rhub_addr=strdup(lst[i]);
				nw.rhub_port=411;
			}
			else
			{
				nw.rhub_port=strtoul(t+1,NULL,10);
				*t='\0';
				nw.rhub_addr=strdup(lst[i]);
			}
			nw.last_checked=0;
			nw.is_up=0;

			rhub_array=g_array_append_val(rhub_array,nw);
			i++;
		}
		
	
		free(vl);
		g_strfreev(lst);
	}
	G_UNLOCK(rhub_array);
	G_UNLOCK(conf_file);
}

/******************************************/
/* send the given string to the given hub */
/******************************************/
static void send_string_to_rhub(RHUB_ENTRY *ptr, GString *up_string)
{
	struct sockaddr_in addr;
	int addr_len;
#ifndef MSG_NOSIGNAL
	void (*old_handler)(int);
#endif

	if(!str_to_inaddr(ptr->rhub_addr,&(addr.sin_addr)))
	{
		addr.sin_family=AF_INET;
		addr.sin_port=htons(ptr->rhub_port);
		addr_len=sizeof(addr);
		
#ifdef MSG_NOSIGNAL
		sendto(multihub_com_socket,up_string->str,MSG_DONTWAIT|MSG_NOSIGNAL,up_string->len,(void*)&addr,addr_len);
#else
		old_handler = signal (SIGPIPE, SIG_IGN);
		sendto(multihub_com_socket,up_string->str,MSG_DONTWAIT,up_string->len,(void*)&addr,addr_len);
		signal (SIGPIPE, old_handler);
#endif
	}
}

/***********************************/
/* check if the remote hubs are up */
/***********************************/
static void	verify_rhub_list(void)
{
	char *v1,*v2;

	G_LOCK(conf_file);
	v1=e_db_str_get(conf_file,"MHUBPASS");
	v2=e_db_str_get(conf_file,"HUBADDR");
	G_UNLOCK(conf_file);

	if( (v1&&(strlen(v1)>5)) &&
		 (v2&&(strlen(v2)>3)) )
	{
		int i;
		time_t cur_time;
		GString *default_string;

		cur_time=time(NULL);
		default_string=g_string_new("");
		g_string_sprintf(default_string,"$Up %s %s|",v1,v2);

		G_LOCK(rhub_array);
		for(i=0;i<rhub_array->len;i++)
		{
			RHUB_ENTRY *ptr;

			ptr=&(g_array_index(rhub_array,RHUB_ENTRY,i));

			if( (ptr->last_checked+RHUB_DELAY)<cur_time)
			{
				send_string_to_rhub(ptr,default_string);
				ptr->is_up=0;							/* the remote hub is marked down until we receive its $UpToo message */
				ptr->last_checked=time(NULL);
			}
		}
		g_string_free(default_string,TRUE);
		G_UNLOCK(rhub_array);
	}
	if(v1!=NULL)
		free(v1);
	if(v2!=NULL)
		free(v2);
}

/*****************************************************/
/* check if the given password is a valid one or not */
/*****************************************************/
/* output: 0=invalid passwd, 1=password ok */
/*******************************************/
static int is_valid_hub_password(char *hpasswd)
{
	char *v1;
	int ret=0;

	if(hpasswd==NULL)
		return 0;

	G_LOCK(conf_file);
	v1=e_db_str_get(conf_file,"MHUBPASS");
	G_UNLOCK(conf_file);

	if(v1==NULL)
		return 0;

	if(!strcmp(v1,hpasswd))
		ret=1;

	free(v1);
	return ret;
}

/******************************************/
/* process the "$Up passwd hubip" message */
/******************************************/
static void h2h_up_process(const char *cmd,int sck,GString *input, char *xtra_param, struct sockaddr_in *raddr, int raddr_len)
{
	char *t;
	char *hpasswd;
	char *hubip;

	t=input->str+strlen(cmd);
	SKIP_SPACE(t);
	if(*t=='\0')
		return ;

	hpasswd=t;
	t=strchr(hpasswd,' ');
	if(t==NULL)
		return;
	*t++='\0';

	SKIP_SPACE(t);
	if(*t=='\0')
		return ;
	hubip=t;
	t=strchr(hubip,'|');
	if(t==NULL)
		return;
	*t='\0';

	if(!is_valid_hub_password(hpasswd))
		return;

	/* mark this hub up and send it a $Uptoo string */
	{
		char *v1,*v2;

		G_LOCK(conf_file);
		v1=e_db_str_get(conf_file,"MHUBPASS");
		v2=e_db_str_get(conf_file,"HUBADDR");
		G_UNLOCK(conf_file);

		if( (v1&&(strlen(v1)>5)) &&
		 	 (v2&&(strlen(v2)>3)) )
		{
			int i;
			G_LOCK(rhub_array);
			for(i=0;i<rhub_array->len;i++)
			{
				RHUB_ENTRY *ptr;
		
				ptr=&(g_array_index(rhub_array,RHUB_ENTRY,i));
				if(!strcmp(hubip,ptr->rhub_addr))
				{
					GString *default_string;

					default_string=g_string_new("");
					g_string_sprintf(default_string,"$UpToo %s %s|",v1,v2);
					send_string_to_rhub(ptr,default_string);
					ptr->is_up=1;							/* the remote hub is marked up */
					ptr->last_checked=time(NULL);
					ptr->laddr=*raddr;
					ptr->laddr_len=raddr_len;
					break;
				}
			}
			G_UNLOCK(rhub_array);
		}

		if(v1!=NULL)
			free(v1);
		if(v2!=NULL)
			free(v2);
	}
}

/*********************************************/
/* process the "$Uptoo passwd hubip" message */
/*********************************************/
static void h2h_uptoo_process(const char *cmd,int sck,GString *input, char *xtra_param, struct sockaddr_in *raddr, int raddr_len)
{
	char *t;
	char *hpasswd;
	char *hubip;

	t=input->str+strlen(cmd);
	SKIP_SPACE(t);
	if(*t=='\0')
		return ;

	hpasswd=t;
	t=strchr(hpasswd,' ');
	if(t==NULL)
		return;
	*t++='\0';

	SKIP_SPACE(t);
	if(*t=='\0')
		return ;
	hubip=t;
	t=strchr(hubip,'|');
	if(t==NULL)
		return;
	*t='\0';

	if(!is_valid_hub_password(hpasswd))
		return;

	/* mark this hub up */
	{
		int i;
		G_LOCK(rhub_array);
		for(i=0;i<rhub_array->len;i++)
		{
			RHUB_ENTRY *ptr;
		
			ptr=&(g_array_index(rhub_array,RHUB_ENTRY,i));
			if(!strcmp(hubip,ptr->rhub_addr))
			{
				ptr->is_up=1;							/* the remote hub is marked up */
				ptr->last_checked=time(NULL);
				ptr->laddr=*raddr;
				ptr->laddr_len=raddr_len;
				break;
			}
		}
		G_UNLOCK(rhub_array);
	}
}

/**************************************/
/* process the "$Search ...." message */
/**************************************/
static void h2h_search_process(const char *cmd,int sck,GString *input, char *xtra_param, struct sockaddr_in *raddr, int raddr_len)
{
	int i;

	/* check if the command comes from a known hub */
	G_LOCK(rhub_array);
	for(i=0;i<rhub_array->len;i++)
	{
		RHUB_ENTRY *ptr;
	
		ptr=&(g_array_index(rhub_array,RHUB_ENTRY,i));
		if((ptr->is_up==1)&&
			(ptr->laddr_len==raddr_len)&&
			(!memcmp(&(ptr->laddr),raddr,raddr_len)) )
		{
			/* known hub found */
			G_UNLOCK(rhub_array);
			queue_const_str_to_everyone(input->str);
			return;
		}
	}
	G_UNLOCK(rhub_array);
}

typedef struct
{
	const char *cmd;
	size_t cmd_len;
	void (*fnc)(const char *cmd,int sck,GString *input, char *xtra_param, struct sockaddr_in *raddr, int raddr_len);
	char *xtra_param;
} CMD_REPLY;

static CMD_REPLY udp_reply[]= {
                                 {"$Up "            , sizeof("$Up ")-1            ,h2h_up_process       ,NULL},
                                 {"$UpToo "         , sizeof("$UpToo ")-1         ,h2h_uptoo_process    ,NULL},
                                 {"$Search "        , sizeof("$Search ")-1        ,h2h_search_process   ,NULL},
                                 {NULL,0,NULL},
                              };

/*********************************************************/
/* this function handles packets coming from remote hubs */
/*********************************************************/
static void manage_hub_to_hub_socket(void)
{
	char buf[8192];
	int ret;
	struct sockaddr_in raddr;
	int raddr_len=sizeof(raddr);

#ifdef MSG_NOSIGNAL
		ret=recvfrom(multihub_com_socket,buf,sizeof(buf),MSG_NOSIGNAL,(void*)&raddr,&raddr_len);
#else
	{
		void (*old_handler)(int);
		old_handler = signal (SIGPIPE, SIG_IGN);
		ret=recvfrom(multihub_com_socket,buf,sizeof(buf),0,(void*)&raddr,&raddr_len);
		signal (SIGPIPE, old_handler);
	}
#endif


	if(ret!=-1)
	{
		int i;
		GString *input;

		buf[ret]='\0';

		input=g_string_new(buf);

		i=0;
		while(udp_reply[i].cmd!=NULL)
		{
			if(!strncmp(input->str,udp_reply[i].cmd,udp_reply[i].cmd_len))
			{
				udp_reply[i].fnc(udp_reply[i].cmd, multihub_com_socket,input, udp_reply[i].xtra_param, &raddr, raddr_len);
				g_string_free(input,TRUE);
				return;
			}
			i++;
		}

		printf("unknown command (2): %s\n",input->str);
		{
			for(i=0;i<input->len;i++)
			{
				printf("%c(%02X) ",input->str[i],((int)input->str[i])&255);
			}

			printf("\n");
		}
		g_string_free(input,TRUE);
	}
	return;
}

/************************************/
/* function dealing with other hubs */
/************************************/
static void manage_hub_to_hub_com(void *ptr)
{
	struct timeval tv;
	fd_set rd;
	int n;

	while(1)
	{
		FD_ZERO(&rd);
		FD_SET(multihub_com_socket,&rd);
		n=multihub_com_socket;

		tv.tv_sec=0;
		tv.tv_usec=100000;

		if(select(n+1,&rd,NULL,NULL,&tv)>0)
		{
			if(FD_ISSET(multihub_com_socket,&rd))
			{
				/* something coming from remote side */
				manage_hub_to_hub_socket();
			}
		}
		verify_rhub_list();
	}
	pthread_exit(NULL);
}

/**********************************************/
/* initialize the multihub communication side */
/**********************************************/
/* output: 0=ok, else !=0 */
/**************************/
int init_multihub_side(void)
{
	static pthread_t thread_id;			/* these 2 variables must exist as long as the thread exists */
	static pthread_attr_t thread_attr;

	printf("int init_multihub_side(void)\n");
	multihub_com_socket=_x_udp(hub_port);
	if(multihub_com_socket==-1)
		return -1;

	/* initialize the array of remote hubs */
	rhub_array=g_array_new(FALSE,FALSE,sizeof(RHUB_ENTRY));
	load_rhub_array();

	/* start the thread dealing with other hubs */
	pthread_attr_init (&thread_attr);
	pthread_attr_setdetachstate(&thread_attr, PTHREAD_CREATE_DETACHED);
	if(pthread_create(&thread_id,&thread_attr,(void*)manage_hub_to_hub_com,NULL)!=0)
	{
		close(multihub_com_socket);
		multihub_com_socket=-1;
		return -1;
	}

	return 0;
}

/***************************************************/
/* send the given string to all active remote hubs */
/***************************************************/
void send_string_to_all_rhubs(GString *str)
{
	printf("send_string_to_all_rhubs: TODO\n");
}

