/* DChub - a Direct Connect Hub for Linux
 * Copyright (C) 2001 Eric Prevoteau
 *
 * multi_private_chat.c: Copyright (C) Eric Prevoteau <www@ac2i.tzo.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>

#include "gvar.h"
#include "cnx_lst.h"
#include "multi_private_chat.h"

typedef struct
{
	char *chat_name;			/* used in the From: field */
	GPtrArray *user_lst;		/* list of nickname of all users of the chat. WARNING: these pointers are the same as the one in CNX_LST->user_nick, not a copy */
									/* this is GString pointer and they should NEVER be freed */
} MCHAT;

static GArray *mchat=NULL;		/* array of MCHAT */
G_LOCK_DEFINE(mchat);

/******************************************/
/* check if the given name is a chat_name */
/******************************************/
/* output: 1=yes, 0=no */
/***********************/
int is_a_mchat_name(const char *name)
{
	int ret=0;

	G_LOCK(mchat);
	if(mchat!=NULL)
	{
		int i;
		for(i=0;i<mchat->len;i++)
		{
			MCHAT *ptr;
	
			ptr=&(g_array_index(mchat,MCHAT,i));
			if((ptr->chat_name!=NULL)&&(!strcmp(name,ptr->chat_name)))
			{
				ret=1;
				break;
			}
		}
	}
	G_UNLOCK(mchat);
	return ret;
}

/*******************************************************/
/* find the index of the given chat in the mchat array */
/*******************************************************/
static int find_chat_idx_by_name(const char *chat_name)
{
	int i;

	if(mchat==NULL)
		return -1;

	for(i=0;i<mchat->len;i++)
	{
		MCHAT *ptr;

		ptr=&(g_array_index(mchat,MCHAT,i));
		if((ptr->chat_name!=NULL)&&(!strcmp(chat_name,ptr->chat_name)))
			return i;
	}

	return -1;
}

/******************************************************/
/* find the index of the given user in the given chat */
/******************************************************/
static int find_user_idx_by_name(const char *user_name, const MCHAT *chat)
{
	int i;

	if(chat==NULL)
		return -1;

	for(i=0;i<chat->user_lst->len;i++)
	{
		GString *str;

		str=g_ptr_array_index(chat->user_lst,i);
		if((str!=NULL)&&(str->str!=NULL)&&(!strcmp(user_name,str->str)))
			return i;
	}

	return -1;
}


/********************************/
/* create a new multi user chat */
/*************************************/
/* cmd syntax: "+new_mchat chatname" */
/******************************************************/
/* cmd == "+new_mchat "                               */
/* ptr is the user making the command                 */
/* dest_base="$To: remotenick From: Hub-Security $"   */
/* str=the full incoming string as it as been entered */
/* xtra_param= NULL                                   */
/* splitted_str= str splitted along " "               */
/******************************************************************************/
/* the function must return >=0 if the connection must remain alive (no error */
/* (0) or recoverable error (>0)) and a value <0 if there is a fatal error)   */
/******************************************************************************/
int create_multi_chat(const char *cmd, CNX_ENTRY *ptr,GString **dest_base, const GString *str, char *xtra_param, const GPtrArray *splitted_str)
{
	int ret=0;
	MCHAT nw;
	GString *output;
	int multi_allow;

	if(splitted_str->len!=2)
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
		send_const_str_to(ptr,(*dest_base)->str);
		return ret;
	}

	G_LOCK(conf_file);
	ret=e_db_int_get(conf_file,"MULTICHAT",&multi_allow);
	G_UNLOCK(conf_file);
	if(ret==0)
	{
		/* multichat is not configured */
		g_string_sprintfa((*dest_base),"MULTICHAT key (int) does not exist, multichat is disabled\r\n|");
		send_const_str_to(ptr,(*dest_base)->str);
		return ret;
	}

	ret=0;

	switch(multi_allow)
	{
		case 0:		/* disabled for everyone */
						return ret;

		case 1:		/* op and master only */
						if (!(ptr->is_op & OP_PRIV))
							return ret;
						break;

		case 2:		/* allowed for everyone */
						break;
	}

	G_LOCK(mchat);
	if(mchat==NULL)
		mchat=g_array_new(FALSE,FALSE,sizeof(MCHAT));

	if (find_chat_idx_by_name(splitted_str->pdata[1]) != -1)
	{
		g_string_sprintfa((*dest_base),"a private chat with the same name already exists.\r\n|");
		send_const_str_to(ptr,(*dest_base)->str);
		goto abrt;
	}

	/* also check if the wanted chat name is not already a user */
	if (find_cnx_by_nickname(splitted_str->pdata[1]))
	{
		g_string_sprintfa((*dest_base),"a user with the same name already exists.\r\n|");
		send_const_str_to(ptr,(*dest_base)->str);
		goto abrt;
	}

	/* and perhaps it is also the special nick "Hub-Security" */
	if(!strcmp("Hub-Security",splitted_str->pdata[1]))
	{
		g_string_sprintfa((*dest_base),"This name is protected, you cannot use it.\r\n|");
		send_const_str_to(ptr,(*dest_base)->str);
		goto abrt;
	}

	nw.chat_name=strdup(splitted_str->pdata[1]);
	if(nw.chat_name==NULL)
	{
		g_string_sprintfa((*dest_base),"Not enought memory to create a new private chat.\r\n|");
		send_const_str_to(ptr,(*dest_base)->str);
		goto abrt;
	}

	nw.user_lst=g_ptr_array_new();
	g_ptr_array_add(nw.user_lst,ptr->user_nick);
	
	mchat=g_array_append_val(mchat,nw);

	/* and send a message to this user on this newly create chat */
	output=g_string_new("");
	g_string_sprintf(output,"$To: %s From: %s $Multi-users chat created\r\n|",ptr->user_nick->str,nw.chat_name);
	send_const_str_to(ptr,output->str);

	g_string_sprintf(output,"$Hello %s|",nw.chat_name);
	send_const_str_to(ptr,output->str);
	g_string_free(output,TRUE);

	abrt:
	G_UNLOCK(mchat);
	return ret;
}

/****************************************************************************************/
/* create "$To: destnick From: chatnick $base_msg" and send it to all users of the chat */
/****************************************************************************************/
static void send_to_users_of_multi_chat_except(MCHAT *m,char *base_msg, int skip_idx)
{
	gchar *tmp;
	int i;

	for(i=0;i<m->user_lst->len;i++)
	{
		GString *dest_user;

		if(i==skip_idx)		/* skip this user */
			continue;

		dest_user=g_ptr_array_index(m->user_lst,i);

		if(dest_user!=NULL)
		{
			tmp=g_strconcat("$To: ",dest_user->str, " From: ",m->chat_name," $",base_msg,NULL);
			send_str_to_named_user(dest_user->str,tmp);
			g_free(tmp);
		}
	}
}

/****************************************************************************/
/* this function is called when a user leaves the hub (on the "quit" event) */
/****************************************************************************/
/* evt_name="quit"            */
/* evt_emitter= user nickname */
/* nb_args= 0                 */
/* xtra_param=NULL            */
/******************************/
void leave_multi_chat(const char *evt_name, const char *evt_emitter, int nb_args, va_list ap, char *xtra_param)
{
	G_LOCK(mchat);
	if(mchat!=NULL)
	{
		int i;
		int j;
		int hit;

		for(i=mchat->len-1;i>=0;i--)		/* from the end to the beginning to avoid error */
		{
			MCHAT *m;

			m=&(g_array_index(mchat,MCHAT,i));
			
			hit=0;

			for(j=m->user_lst->len-1;j>=0;j--)
			{
				GString *s;

				s=g_ptr_array_index(m->user_lst,j);
				if((s!=NULL)&&(!strcmp(s->str,evt_emitter)))
				{
					g_ptr_array_remove_index_fast(m->user_lst,j);	/* NEVER free s */
					hit=1;
				}
			}

			if(hit)
			{
				gchar *tmp;

				tmp=g_strconcat(evt_emitter," has left the chat.\r\n|",NULL);
				send_to_users_of_multi_chat_except(m,tmp,-1);
				g_free(tmp);

				if(m->user_lst->len==0)
				{
					/* there is no one on this chat, destroy it */
					free(m->chat_name);
					g_ptr_array_free(m->user_lst,TRUE);
					mchat=g_array_remove_index_fast(mchat,i);
				}
			}
		}
	}
	G_UNLOCK(mchat);
}

/************************************************************************************/
/* this function is called when a user sends a private message (= "privchat" event) */
/************************************************************************************/
/* evt_name="privchat"                */
/* evt_emitter= user nickname         */
/* nb_args= 2                         */
/* ap={"dest_nick","message content"} */
/* xtra_param=NULL                    */
/**************************************/
void privchat_multi_chat(const char *evt_name, const char *evt_emitter, int nb_args, va_list ap, char *xtra_param)
{
	int idx;
	int uidx;
	char *dest_nick;
	char *msg_content;
	MCHAT *m;

	dest_nick=va_arg(ap,char *);	/* it is in fact the chat nickname */
	msg_content=va_arg(ap,char *);

	G_LOCK(mchat);
	idx=find_chat_idx_by_name(dest_nick);
	if(idx!=-1)
	{
		m=&(g_array_index(mchat,MCHAT,idx));
		uidx=find_user_idx_by_name(evt_emitter,m);		/* is this user allowed in this chat ? */
		if(uidx!=-1)
		{
			gchar *tmp;

			tmp=g_strconcat(msg_content,"|",NULL);
			send_to_users_of_multi_chat_except(m,tmp,uidx);	/* send the message to all users except the one sending the message */
			g_free(tmp);
		}
	}
	G_UNLOCK(mchat);
}

/***************************************************************************************/
/* this function is called when a user retrieves the user list (= "getnicklist" event) */
/***************************************************************************************/
/* evt_name="getnicklist"             */
/* evt_emitter= user nickname         */
/* nb_args= 0                         */
/* xtra_param=NULL                    */
/**************************************/
void getnicklist_multi_chat(const char *evt_name, const char *evt_emitter, int nb_args, va_list ap, char *xtra_param)
{
	int i;
	int uidx;
	MCHAT *m;
	GString *output;

	G_LOCK(mchat);
	if(mchat!=NULL)
	{
		output=g_string_new("");

		for(i=0;i<mchat->len;i++)
		{
			m=&(g_array_index(mchat,MCHAT,i));

			uidx=find_user_idx_by_name(evt_emitter,m);	/* is this user on this chat ? */
			if(uidx!=-1)
			{
				/* yes ? send user+ message with the name of the chat */
				g_string_sprintf(output,"$Hello %s|",m->chat_name);
				send_str_to_named_user(evt_emitter,output->str);
			}
		}
		g_string_free(output,TRUE);
	}
	G_UNLOCK(mchat);
}


/*******************************************************/
/* invite the given users to the given multi user chat */
/****************************************************************/
/* cmd syntax: "+invite_mchat chatname nickname [nickname ...]" */
/****************************************************************/
/* cmd == "+invite_mchat "                            */
/* ptr is the user making the command                 */
/* dest_base="$To: remotenick From: Hub-Security $"   */
/* str=the full incoming string as it as been entered */
/* xtra_param= NULL                                   */
/* splitted_str= str splitted along " "               */
/******************************************************************************/
/* the function must return >=0 if the connection must remain alive (no error */
/* (0) or recoverable error (>0)) and a value <0 if there is a fatal error)   */
/******************************************************************************/
int invite_multi_chat(const char *cmd, CNX_ENTRY *ptr,GString **dest_base, const GString *str, char *xtra_param, const GPtrArray *splitted_str)
{
	int ret=0;
	int idx;
	int uidx;
	MCHAT *chat;
	int hit=0;

	if(splitted_str->len<3)
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s. Mini: 3\r\n|",cmd);
		send_const_str_to(ptr,(*dest_base)->str);
		return ret;
	}

	G_LOCK(mchat);
	idx=find_chat_idx_by_name(splitted_str->pdata[1]);
	if(idx==-1)
	{
		g_string_sprintfa((*dest_base),"no chat with this name exists\r\n|");
		send_const_str_to(ptr,(*dest_base)->str);
		goto abrt;
	}

	chat=&(g_array_index(mchat,MCHAT,idx));
	uidx=find_user_idx_by_name(ptr->user_nick->str,chat);		/* is this user member of this chat ? */
	if(uidx==-1)
	{
		g_string_sprintfa((*dest_base),"you are not member of this chat\r\n|");
		send_const_str_to(ptr,(*dest_base)->str);
		goto abrt;
	}

	{
		CNX_ENTRY *new_user;
		GString *output;
		int i;

		output=g_string_new("");
		for(i=2;i<splitted_str->len;i++)	/* skip the command and the chatname */
		{
			new_user=find_cnx_by_nickname(splitted_str->pdata[i]);
			if(new_user==NULL)
			{
				/* no user with this name on the hub */
				g_string_sprintf(output,"$To: %s From: %s $No user named %s is on the hub\r\n|",ptr->user_nick->str,chat->chat_name,(char*)(splitted_str->pdata[i]));
				send_const_str_to(ptr,output->str);
			}
			else
			{
				/* a user with this name exists but perhaps he is already on the chat */
				int uidx2;

				uidx2=find_user_idx_by_name(splitted_str->pdata[i],chat);
				if(uidx2==-1)
				{
					g_ptr_array_add(chat->user_lst,new_user->user_nick);		/* add the user to the user list */
					hit=1;
				}

				/* force the "fake" user to be redisplayed */
				g_string_sprintf(output,"$Hello %s|",chat->chat_name);
				send_str_to_named_user(splitted_str->pdata[i],output->str);

				g_string_sprintf(output,"$To: %s From: %s $You are invited to a private multi users chat\r\n|",(char*)(splitted_str->pdata[i]),chat->chat_name);
				send_str_to_named_user(splitted_str->pdata[i],output->str);
			}
		}
		g_string_free(output,TRUE);
	}

	if(hit)
	{
		GString *output;
		int i;

		output=g_string_new("chat user list: ");
		for(i=0;i<chat->user_lst->len;i++)
		{
			if(i!=0)
			{
				output=g_string_append(output," , ");
			}
			output=g_string_append(output,((GString*)(g_ptr_array_index(chat->user_lst,i)))->str);
		}
		output=g_string_append(output,"\r\n|");
		send_to_users_of_multi_chat_except(chat,output->str,-1);
		g_string_free(output,TRUE);
	}

	/* and send a message to this user on this newly create chat */

	abrt:
	G_UNLOCK(mchat);
	return ret;
}

/***********************************/
/* leave the given multi user chat */
/**************************************/
/* cmd syntax: "+leave_mchat chatname */
/******************************************************/
/* cmd == "+leave_mchat "                             */
/* ptr is the user making the command                 */
/* dest_base="$To: remotenick From: Hub-Security $"   */
/* str=the full incoming string as it as been entered */
/* xtra_param= NULL                                   */
/* splitted_str= str splitted along " "               */
/******************************************************************************/
/* the function must return >=0 if the connection must remain alive (no error */
/* (0) or recoverable error (>0)) and a value <0 if there is a fatal error)   */
/******************************************************************************/
int cmd_leave_multi_chat(const char *cmd, CNX_ENTRY *ptr,GString **dest_base, const GString *str, char *xtra_param, const GPtrArray *splitted_str)
{
	int ret=0;
	int idx;
	int uidx;
	MCHAT *chat;

	if(splitted_str->len!=2)	/* cmd + 1 param */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s. Must be 1\r\n|",cmd);
		send_const_str_to(ptr,(*dest_base)->str);
		return ret;
	}

	G_LOCK(mchat);
	idx=find_chat_idx_by_name(splitted_str->pdata[1]);
	if(idx==-1)
	{
		g_string_sprintfa((*dest_base),"no chat with this name exists\r\n|");
		send_const_str_to(ptr,(*dest_base)->str);
		goto abrt;
	}

	chat=&(g_array_index(mchat,MCHAT,idx));
	uidx=find_user_idx_by_name(ptr->user_nick->str,chat);
	if(uidx==-1)
	{
		g_string_sprintfa((*dest_base),"you are not member of this chat\r\n|");
		send_const_str_to(ptr,(*dest_base)->str);
		goto abrt;
	}

	/* remove the user from this chat */
	g_ptr_array_remove_index_fast(chat->user_lst,uidx);		/* NEVER free s */

	/* and notify this change to other user */
	{
		gchar *tmp;

		tmp=g_strconcat(ptr->user_nick->str," has left the chat.\r\n|",NULL);
		send_to_users_of_multi_chat_except(chat,tmp,-1);
		g_free(tmp);
	}

	/* force the "fake" user to be undisplayed */
	{
		GString *output;
		output=g_string_new("");
		g_string_sprintf(output,"$Quit %s|",chat->chat_name);
		send_const_str_to(ptr,output->str);

		g_string_sprintf(output,"$To: %s From: %s $You have left this chat.\r\n|",ptr->user_nick->str,chat->chat_name);
		send_const_str_to(ptr,output->str);
		g_string_free(output,TRUE);
	}

	/* and delete the chat if it becomes empty */
	if(chat->user_lst->len==0)
	{
		/* there is no one on this chat, destroy it */
		free(chat->chat_name);
		g_ptr_array_free(chat->user_lst,TRUE);
		mchat=g_array_remove_index_fast(mchat,idx);
	}

	abrt:
	G_UNLOCK(mchat);
	return ret;
}


